from abc import abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Iterator, Sequence

import numpy as np


class ValueType(str, Enum):
    """Specifies supported value types supported by Variables."""

    REAL = "real"
    SIGMOID = "sigmoid"
    PROBABILITY = "probability"  # single probability
    PROBABILITY_SAMPLE = "probability_sample"  # sample from single probability
    PROBABILITY_DISTRIBUTION = (
        "probability_distribution"  # probability distribution over classes
    )
    CLASS = "class"  # sample from probability distribution over classes


class Variable:
    """Describes an experimental variable: name, type, range, units, and value of a variable."""

    def __init__(
        self,
        name="",
        value_range=(0, 1),
        units="",
        type=ValueType.REAL,
        variable_label="",
        rescale=1,
        is_covariate=False,
        participant=None,
    ):

        self._name = name
        self._units = units
        self._value_range = value_range
        self._value = 0
        self.type = type
        if variable_label == "":
            self._variable_label = self._name
        else:
            self._variable_label = variable_label
        self._rescale = rescale
        self._is_covariate = is_covariate
        self._participant = participant

    def __get_value_range__(self):
        """Get range of variable.
        The variable range determines the minimum and maximum allowed value
        to be manipulated or measured."""
        return self._value_range

    def __set_value_range__(self, value_range):
        """Set range of variable.
        The variable range determines the minimum and maximum allowed value
        to be manipulated or measured."""
        self._value_range = value_range

    def __cap_value__(self, value):
        """Cap value of variable."""
        minimum = self._value_range[0]
        maximum = self._value_range[1]
        return np.min([np.max([value, minimum]), maximum])

    def get_value(self):
        """Get value."""
        return self._value * self._rescale

    def set_value(self, value):
        """Set value."""
        self._value = self.__cap_value__(value)

    def get_value_from_dict(self, dictionary, position):
        """Reads value of independent variable from a dictionary."""

        value_list = dictionary[self.get_name()]

        if position > len(value_list):
            raise Exception(
                f"Queried position "
                f"{str(position)}"
                f" for variable "
                f"{self.get_name()}"
                f"'exceeds number of available positions for that variable in the dictionary."
            )

        return value_list[position] * self._rescale

    def get_value_list_from_dict(self, dictionary):
        """Gets the rescaled values of independent variables from a dictionary."""
        value_list = dictionary[self.get_name()]

        rescaled_list = [element * self._rescale for element in value_list]

        return rescaled_list

    def set_value_from_dict(self, dictionary, position):
        """
        Reads and sets value of independent variable from a dictionary
        with variable_label being the key
        """

        value_list = dictionary[self.get_name()]

        if position > len(value_list):
            raise Exception(
                f"Queried position "
                f"{str(position)}"
                f" for variable "
                f"{self.get_name()}"
                f" exceeds number of available positions for that variable in the dictionary."
            )

        self.set_value(value_list[position])

    def get_name(self):
        """Get variable name."""
        return self._name

    def set_name(self, name):
        """Set variable name."""
        self._name = name

    def get_units(self):
        """Get variable units."""
        return self._units

    def set_units(self, units):
        """Set variable units."""
        self._units = units

    def get_variable_label(self):
        """Get variable label."""
        return self._variable_label

    def set_variable_label(self, variable_label):
        """Set variable label."""
        self._variable_label = variable_label

    def set_covariate(self, is_covariate):
        """Set whether this dependent variable is treated as covariate."""
        self._is_covariate = is_covariate


@dataclass(frozen=True)
class VariableCollection:
    """Immutable metadata about dependent / independent variables and covariates."""

    independent_variables: Sequence[Variable]
    dependent_variables: Sequence[Variable]
    covariates: Sequence[Variable] = field(default_factory=list)

    @property
    def all_variables(self) -> Iterator[Variable]:
        for vars in (
            self.independent_variables,
            self.dependent_variables,
            self.covariates,
        ):
            for v in vars:
                yield v

    @property
    def variable_names(self):
        return (v.get_name() for v in self.all_variables)

    @property
    def output_type(self):
        first_type = self.dependent_variables[0].type
        assert all(dv.type == first_type for dv in self.dependent_variables), (
            "Dependent variable output types don't match. "
            "Different output types are not supported yet."
        )
        return first_type

    @property
    def input_dimensions(self):
        """The number of independent variables and covariates."""
        return len(self.independent_variables) + len(self.covariates)

    @property
    def output_dimensions(self):
        """The number of dependent variables."""
        return len(self.dependent_variables)


class IV(Variable):
    def __init__(self, *args, **kwargs):
        self._name = "IV"
        self._variable_label = "Independent Variable"

        super().__init__(*args, **kwargs)

    # Method for measuring dependent variable.
    @abstractmethod
    def manipulate(self):
        pass


class DV(Variable):
    def __init__(self, *args, **kwargs):
        self._name = "DV"
        self._variable_label = "Dependent Variable"

        self._is_covariate = False

        super().__init__(*args, **kwargs)

    # Method for measuring dependent variable.
    @abstractmethod
    def measure(self):
        pass

    # Get whether this dependent variable is treated as covariate.
    def __is_covariate__(self):
        return self._is_covariate

    # Set whether this dependent variable is treated as covariate.
    def __set_covariate__(self, is_covariate):
        self._is_covariate = is_covariate


class IVInSilico(IV):
    _variable_label = "IV"
    _name = "independent variable"
    _units = "activation"
    _priority = 0
    _value_range = (0, 1)
    _value = 0

    def __init__(self, *args, **kwargs):
        super(IVInSilico, self).__init__(*args, **kwargs)

    def manipulate(self):
        self._participant.set_value(self._name, self.get_value())


class DVInSilico(DV):
    _variable_label = "DV"
    _name = "dependent variable"
    _units = "activation"
    _priority = 0
    _value_range = (0, 1)
    _value = 0

    def __init__(self, *args, **kwargs):
        super(DVInSilico, self).__init__(*args, **kwargs)

    def measure(self):
        measurement = self._participant.get_value(self._name)
        self.set_value(measurement)


class IVTrial(IV):

    _name = "trial"
    _UID = ""
    _variable_label = "Trial"
    _units = "trials"
    _priority = 0
    _value_range = (0, 10000000)
    _value = 0

    def __init__(self, *args, **kwargs):
        super(IVTrial, self).__init__(*args, **kwargs)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):
        pass

    def disconnect(self):
        pass



from autora.variable import DVInSilico, IVInSilico, IVTrial
from autora.variable.time import DVTime, IVTime
from autora.variable.tinkerforge import DVCurrent, DVVoltage, IVCurrent, IVVoltage

dv_labels = {
    "time_DV": (DVTime, "Time", "", "time_DV", "s", 0, (0, 3600)),
    "voltage0": (DVVoltage, "Voltage 0", "MjY", "voltage0", "mV", 1, (-3500, 3500)),
    "voltage1": (DVVoltage, "Voltage 1", "MjY", "voltage1", "mV", 1, (-3500, 3500)),
    "current0": (DVCurrent, "Current 0", "Hfg", "current0", "mA", 2, (0, 20)),
    "current1": (DVCurrent, "Current 1", "Hfg", "current1", "mA", 2, (0, 20)),
    "verbal_red": (
        DVInSilico,
        "Verbal Response Red",
        None,
        "verbal_red",
        "activation",
        0,
        (0, 1),
    ),
    "verbal_green": (
        DVInSilico,
        "Verbal Response Green",
        None,
        "verbal_green",
        "activation",
        0,
        (0, 1),
    ),
    "verbal_sample": (
        DVInSilico,
        "Verbal Response Sample",
        None,
        "verbal_sample",
        "class",
        0,
        (0, 1),
    ),
    "difference_detected": (
        DVInSilico,
        "Difference Detected",
        None,
        "difference_detected",
        "activation",
        0,
        (0, 1),
    ),
    "difference_detected_sample": (
        DVInSilico,
        "Difference Detected",
        None,
        "difference_detected_sample",
        "class",
        0,
        (0, 1),
    ),
    "learning_performance": (
        DVInSilico,
        "Accuracy",
        None,
        "learning_performance",
        "probability",
        0,
        (0, 1),
    ),
    "learning_performance_sample": (
        DVInSilico,
        "Accuracy Sample",
        None,
        "learning_performance_sample",
        "class",
        0,
        (0, 1),
    ),
    "dx1_lca": (
        DVInSilico,
        "dx1",
        None,
        "dx1_lca",
        "net input delta",
        0,
        (-1000, 1000),
    ),
}
iv_labels = {
    "time_IV": (IVTime, "Time", "", "time_IV", "s", 1, (0, 3600)),
    "trial": (IVTrial, "Trial", "", "trial", "trials", 0, (0, 10000000)),
    "source_voltage": (
        IVVoltage,
        "Source Voltage",
        "MST",
        "source_voltage",
        "mV",
        2,
        (0, 5000),
    ),
    "source_current": (
        IVCurrent,
        "Source Current",
        "MST",
        "source_current",
        "µA",
        2,
        (0, 20000),
    ),
    "color_red": (
        IVInSilico,
        "Color Unit Red",
        None,
        "color_red",
        "activation",
        0,
        (0, 1),
    ),
    "color_green": (
        IVInSilico,
        "Color Unit Green",
        None,
        "color_green",
        "activation",
        0,
        (0, 1),
    ),
    "word_red": (
        IVInSilico,
        "Word Unit Red",
        None,
        "word_red",
        "activation",
        0,
        (0, 1),
    ),
    "word_green": (
        IVInSilico,
        "Word Unit Green",
        None,
        "word_green",
        "activation",
        0,
        (0, 1),
    ),
    "task_color": (
        IVInSilico,
        "Task Unit Color Naming",
        None,
        "task_color",
        "activation",
        0,
        (0, 1),
    ),
    "task_word": (
        IVInSilico,
        "Task Unit Word Reading",
        None,
        "task_word",
        "activation",
        0,
        (0, 1),
    ),
    "S1": (IVInSilico, "Stimulus 1 Intensity", None, "S1", "activation", 0, (0, 5)),
    "S2": (IVInSilico, "Stimulus 2 Intensity", None, "S2", "activation", 0, (0, 5)),
    "learning_trial": (
        IVInSilico,
        "Trial",
        None,
        "learning_trial",
        "trial",
        0,
        (0, 1000),
    ),
    "P_initial": (
        IVInSilico,
        "Initial Performance",
        None,
        "P_initial",
        "probability",
        0,
        (0, 1),
    ),
    "P_asymptotic": (
        IVInSilico,
        "Best Performance",
        None,
        "P_asymptotic",
        "probability",
        0,
        (0, 1),
    ),
    "x1_lca": (IVInSilico, "x1", None, "x1_lca", "net input", 0, (-1000, 1000)),
    "x2_lca": (IVInSilico, "x2", None, "x2_lca", "net input", 0, (-1000, 1000)),
    "x3_lca": (IVInSilico, "x3", None, "x3_lca", "net input", 0, (-1000, 1000)),
}



import copy
import logging
from dataclasses import dataclass
from functools import partial
from itertools import cycle
from typing import Callable, Iterator, Literal, Optional, Sequence

import numpy as np
import torch
import torch.nn
import torch.nn.utils
import torch.utils.data
import tqdm
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

from autora.theorist.darts.architect import Architect
from autora.theorist.darts.dataset import darts_dataset_from_ndarray
from autora.theorist.darts.model_search import DARTSType, Network
from autora.theorist.darts.utils import (
    AvgrageMeter,
    format_input_target,
    get_loss_function,
    get_output_format,
    get_output_str,
)
from autora.theorist.darts.visualize import darts_model_plot
from autora.variable import ValueType

_logger = logging.getLogger(__name__)

SAMPLING_STRATEGIES = Literal["max", "sample"]

progress_indicator = tqdm.auto.tqdm


@dataclass(frozen=True)
class _DARTSResult:
    """A container for passing fitted DARTS results around."""

    network_: Network
    model_: torch.nn.Module
    model_sampler_: Callable[[SAMPLING_STRATEGIES], torch.nn.Module]


def _general_darts(
    X: np.ndarray,
    y: np.ndarray,
    batch_size: int = 20,
    num_graph_nodes: int = 2,
    output_type: ValueType = ValueType.REAL,
    classifier_weight_decay: float = 1e-2,
    darts_type: DARTSType = DARTSType.ORIGINAL,
    init_weights_function: Optional[Callable] = None,
    learning_rate: float = 2.5e-2,
    learning_rate_min: float = 0.01,
    momentum: float = 9e-1,
    optimizer_weight_decay: float = 3e-4,
    param_updates_per_epoch: int = 20,
    arch_updates_per_epoch: int = 20,
    arch_weight_decay: float = 1e-4,
    arch_weight_decay_df: float = 3e-4,
    arch_weight_decay_base: float = 0.0,
    arch_learning_rate: float = 3e-3,
    fair_darts_loss_weight: int = 1,
    max_epochs: int = 100,
    grad_clip: float = 5,
) -> _DARTSResult:
    """
    Function to implement the DARTS optimization, given a fixed architecture and input data.
    """

    _logger.info("Starting fit initialization")

    data_loader, input_dimensions, output_dimensions = _get_data_loader(
        X=X,
        y=y,
        batch_size=batch_size,
    )

    criterion = get_loss_function(output_type)
    output_function = get_output_format(output_type)

    network_ = Network(
        num_classes=output_dimensions,
        criterion=criterion,
        steps=num_graph_nodes,
        n_input_states=input_dimensions,
        classifier_weight_decay=classifier_weight_decay,
        darts_type=darts_type,
    )
    if init_weights_function is not None:
        network_.apply(init_weights_function)

    # Generate the architecture of the model
    architect = Architect(
        network_,
        momentum=momentum,
        arch_weight_decay=arch_weight_decay,
        arch_weight_decay_df=arch_weight_decay_df,
        arch_weight_decay_base=arch_weight_decay_base,
        fair_darts_loss_weight=fair_darts_loss_weight,
        arch_learning_rate=arch_learning_rate,
    )

    optimizer = torch.optim.SGD(
        params=network_.parameters(),
        lr=learning_rate,
        momentum=momentum,
        weight_decay=optimizer_weight_decay,
    )

    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
        optimizer=optimizer,
        T_max=param_updates_per_epoch,
        eta_min=learning_rate_min,
    )

    coefficient_optimizer = partial(
        _optimize_coefficients,
        criterion=criterion,
        data_loader=data_loader,
        grad_clip=grad_clip,
        optimizer=optimizer,
        param_updates_per_epoch=param_updates_per_epoch,
        scheduler=scheduler,
    )

    _logger.info("Starting fit.")
    network_.train()

    for epoch in progress_indicator(range(max_epochs)):

        _logger.debug(f"Running fit, epoch {epoch}")

        # Do the Architecture update

        # First reset the data iterator
        data_iterator = _get_data_iterator(data_loader)

        # Then run the arch optimization
        for arch_step in range(arch_updates_per_epoch):
            _logger.debug(
                f"Running architecture update, "
                f"epoch: {epoch}, architecture: {arch_step}"
            )

            X_batch, y_batch = _get_next_input_target(
                data_iterator, criterion=criterion
            )

            architect.step(
                input_valid=X_batch,
                target_valid=y_batch,
                network_optimizer=optimizer,
                unrolled=False,
            )

        # Then run the param optimization
        coefficient_optimizer(network_)

    model_ = _generate_model(
        network_=network_,
        coefficient_optimizer=coefficient_optimizer,
        output_function=output_function,
        sampling_strategy="max",
    )
    model_sampler_ = partial(
        _generate_model,
        network_=network_,
        coefficient_optimizer=coefficient_optimizer,
        output_function=output_function,
    )

    results = _DARTSResult(
        model_=model_, model_sampler_=model_sampler_, network_=network_
    )

    return results


def _optimize_coefficients(
    network: Network,
    criterion: Callable,
    data_loader: torch.utils.data.DataLoader,
    grad_clip: bool,
    optimizer: torch.optim.Optimizer,
    param_updates_per_epoch: int,
    scheduler: torch.optim.lr_scheduler.CosineAnnealingLR,
):
    """
    Function to optimize the coefficients of a DARTS Network.

    Warning: This modifies the coefficients of the Network in place.
    """

    data_iterator = _get_data_iterator(data_loader)

    objs = AvgrageMeter()

    for param_step in range(param_updates_per_epoch):
        _logger.debug(f"Running parameter update, " f"param: {param_step}")

        lr = scheduler.get_last_lr()[0]
        X_batch, y_batch = _get_next_input_target(data_iterator, criterion=criterion)
        optimizer.zero_grad()

        # compute loss for the model
        logits = network(X_batch)
        loss = criterion(logits, y_batch)

        # update gradients for model
        loss.backward()

        # clips the gradient norm
        torch.nn.utils.clip_grad_norm_(network.parameters(), grad_clip)

        # moves optimizer one step (applies gradients to weights)
        optimizer.step()

        # applies weight decay to classifier weights
        network.apply_weight_decay_to_classifier(lr)

        # moves the annealing scheduler forward to determine new learning rate
        scheduler.step()

        # compute accuracy metrics
        n = X_batch.size(0)
        objs.update(loss.data, n)


def _get_data_loader(
    X: np.ndarray,
    y: np.ndarray,
    batch_size: int,
) -> torch.utils.data.DataLoader:
    """Construct a minimal torch.utils.data.DataLoader for the input data."""

    X_, y_ = check_X_y(X, y, ensure_2d=True, multi_output=True)

    if y_.ndim == 1:
        y_ = y_.reshape((y_.size, 1))

    input_dimensions = X_.shape[1]
    output_dimensions = y_.shape[1]

    experimental_data = darts_dataset_from_ndarray(X_, y_)

    data_loader = torch.utils.data.DataLoader(
        experimental_data,
        batch_size=batch_size,
        shuffle=True,
        pin_memory=True,
        num_workers=0,
    )
    return data_loader, input_dimensions, output_dimensions


def _get_data_iterator(data_loader: torch.utils.data.DataLoader) -> Iterator:
    data_iterator = cycle(iter(data_loader))
    return data_iterator


def _get_next_input_target(data_iterator: Iterator, criterion: torch.nn.Module):
    input_search, target_search = next(data_iterator)

    input_var = torch.autograd.Variable(input_search, requires_grad=False)
    target_var = torch.autograd.Variable(target_search, requires_grad=False)

    input_fmt, target_fmt = format_input_target(
        input_var, target_var, criterion=criterion
    )
    return input_fmt, target_fmt


def _generate_model(
    sampling_strategy: SAMPLING_STRATEGIES,
    network_: Network,
    coefficient_optimizer: Callable[[Network], None],
    output_function: torch.nn.Module,
):

    # Set edges in the network with the highest weights to 1, others to 0
    model_without_output_function = copy.deepcopy(network_)

    if sampling_strategy == "max":
        new_weights = model_without_output_function.max_alphas_normal()
    elif sampling_strategy == "sample":
        new_weights = model_without_output_function.sample_alphas_normal()

    model_without_output_function.fix_architecture(True, new_weights=new_weights)

    # Re-optimize the parameters
    coefficient_optimizer(model_without_output_function)

    # Include the output function
    model_ = torch.nn.Sequential(model_without_output_function, output_function)

    return model_


class DARTSRegressor(BaseEstimator, RegressorMixin):
    """
    Differentiable ARchiTecture Search Regressor.

    DARTS finds a composition of functions and coefficients to minimize a loss function suitable for
    the dependent variable.

    This class is intended to be compatible with the
    [Scikit-Learn Estimator API](https://scikit-learn.org/stable/developers/develop.html).

    Examples:

        >>> import numpy as np
        >>> num_samples = 1000
        >>> X = np.linspace(start=0, stop=1, num=num_samples).reshape(-1, 1)
        >>> y = 15. * np.ones(num_samples)
        >>> estimator = DARTSRegressor(num_graph_nodes=1)
        >>> estimator = estimator.fit(X, y)
        >>> estimator.predict([[0.5]])
        array([[15.051043]], dtype=float32)


    Attributes:
        network_: represents the optimized network for the architecture search, without the
            output function
        model_: represents the best-fit model including the output function
            after sampling of the network to pick a single computation graph.
            By default, this is the computation graph with the maximum weights,
            but can be set to a graph based on a sample on the edge weights
            by running the `resample_model(sample_strategy="sample")` method.
            It can be reset by running the `resample_model(sample_strategy="max")` method.
        model_sampler_: a callable which generates versions of the model, either based on the
            maximum architecture weights or based on a sample over the architecture weights.



    """

    def __init__(
        self,
        batch_size: int = 64,
        num_graph_nodes: int = 2,
        classifier_weight_decay: float = 1e-2,
        darts_type: DARTSType = DARTSType.ORIGINAL,
        init_weights_function: Optional[Callable] = None,
        learning_rate: float = 2.5e-2,
        learning_rate_min: float = 0.01,
        momentum: float = 9e-1,
        optimizer_weight_decay: float = 3e-4,
        param_updates_per_epoch: int = 10,
        arch_updates_per_epoch: int = 1,
        arch_weight_decay: float = 1e-4,
        arch_weight_decay_df: float = 3e-4,
        arch_weight_decay_base: float = 0.0,
        arch_learning_rate: float = 3e-3,
        fair_darts_loss_weight: int = 1,
        max_epochs: int = 10,
        grad_clip: float = 5,
        output_type: ValueType = ValueType.REAL,
    ) -> None:
        """
        Arguments:
            batch_size: number of observations to be used per update
            num_graph_nodes: number of intermediate nodes in the DARTS graph.
            classifier_weight_decay:
            darts_type:
            init_weights_function:
            learning_rate:
            learning_rate_min:
            momentum:
            optimizer_weight_decay:
            param_updates_per_epoch:
            arch_updates_per_epoch:
            arch_weight_decay:
            arch_weight_decay_df:
            arch_weight_decay_base:
            arch_learning_rate:
            fair_darts_loss_weight:
            max_epochs:
            grad_clip:
        """

        self.batch_size = batch_size

        self.num_graph_nodes = num_graph_nodes
        self.classifier_weight_decay = classifier_weight_decay
        self.darts_type = darts_type
        self.init_weights_function = init_weights_function

        self.learning_rate = learning_rate
        self.learning_rate_min = learning_rate_min
        self.momentum = momentum
        self.optimizer_weight_decay = optimizer_weight_decay

        self.param_updates_per_epoch = param_updates_per_epoch

        self.arch_updates_per_epoch = arch_updates_per_epoch
        self.arch_weight_decay = arch_weight_decay
        self.arch_weight_decay_df = arch_weight_decay_df
        self.arch_weight_decay_base = arch_weight_decay_base
        self.arch_learning_rate = arch_learning_rate
        self.fair_darts_loss_weight = fair_darts_loss_weight

        self.max_epochs = max_epochs
        self.grad_clip = grad_clip

        self.output_type = output_type

        self.X_: Optional[np.ndarray] = None
        self.y_: Optional[np.ndarray] = None
        self.network_: Optional[Network] = None
        self.model_: Optional[Network] = None
        self.model_sampler_: Optional[Callable[[SAMPLING_STRATEGIES], Network]] = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        """
        Runs the optimization for a given set of `X`s and `y`s.

        Arguments:
            X: independent variables in an n-dimensional array
            y: dependent variables in an n-dimensional array

        Returns:
            self (DARTSRegressor): the fitted estimator
        """
        params = self.get_params()

        if self.darts_type == ValueType.CLASS:
            raise NotImplementedError(
                "Classification not implemented for DARTSRegressor."
            )

        fit_results = _general_darts(X=X, y=y, **params)
        self.X_ = X
        self.y_ = y
        self.network_ = fit_results.network_
        self.model_ = fit_results.model_
        self.model_sampler_ = fit_results.model_sampler_
        return self

    def resample_model(
        self, sampling_strategy: SAMPLING_STRATEGIES = "sample"
    ) -> Network:
        """
        Generates a new model based on a sample of the architecture weights
        (`sampling_strategy="sample"`) or the maximum architecture weights
        (`sampling_strategy="max"`)

        Args:
            sampling_strategy:

        Returns:

        """
        check_is_fitted(self, attributes=["model_sampler_"])
        assert self.model_sampler_ is not None
        self.model_ = self.model_sampler_(sampling_strategy)
        return self.model_

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Applies the fitted model to a set of independent variables `X`,
        to give predictions for the dependent variable `y`.

        Arguments:
            X: independent variables in an n-dimensional array

        Returns:
            y: predicted dependent variable values
        """
        X_ = check_array(X)

        # First run the checks using the scikit-learn API, listing the key parameters
        check_is_fitted(self, attributes=["model_"])

        # Since self.model_ is initialized as None, mypy throws an error if we
        # just call self.model_(X) in the predict method, as it could still be none.
        # MyPy doesn't understand that the sklearn check_is_fitted function
        # ensures the self.model_ parameter is initialized and otherwise throws an error,
        # so we check that explicitly here and pass the model which can't be None.
        assert self.model_ is not None

        y_ = self.model_(torch.as_tensor(X_).float())
        y = y_.detach().numpy()

        return y

    def visualize_model(
        self,
        input_labels: Optional[Sequence[str]] = None,
    ):
        assert self.model_ is not None
        fitted_sampled_network = self.model_[0]

        genotype = Network.genotype(fitted_sampled_network).normal
        (
            _,
            _,
            param_list,
        ) = fitted_sampled_network.countParameters()

        assert self.X_ is not None

        if input_labels is not None:
            input_labels_ = tuple(input_labels)
        elif hasattr(self.X_, "columns"):
            input_labels_ = tuple(self.X_.columns)
        else:
            input_dim = 1 if self.X_.ndim == 1 else self.X_.shape[1]
            input_labels_ = tuple(f"x{i}" for i in range(input_dim))

        assert self.y_ is not None
        out_dim = 1 if self.y_.ndim == 1 else self.y_.shape[1]
        out_func = get_output_str(self.output_type)

        # call to plot function
        graph = darts_model_plot(
            genotype=genotype,
            input_labels=input_labels_,
            param_list=param_list,
            full_label=True,
            out_dim=out_dim,
            out_fnc=out_func,
        )

        return graph



import time

from autora.variable import DV, IV


class VTime:

    _t0 = 0

    def __init__(self):
        self._t0 = time.time()

    # Resets reference time.
    def reset(self):
        self._t0 = time.time()


class IVTime(IV, VTime):

    _name = "time_IV"
    _UID = ""
    _variable_label = "Time"
    _units = "s"
    _priority = 0
    _value_range = (0, 3600)
    _value = 0

    # Initializes reference time.
    # The reference time usually denotes the beginning of an experiment trial.
    def __init__(self, *args, **kwargs):
        super(IVTime, self).__init__(*args, **kwargs)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):

        t_wait = self.get_value() - (time.time() - self._t0)
        if t_wait <= 0:
            return
        else:
            time.sleep(t_wait)

    def disconnect(self):
        pass


class DVTime(DV, VTime):

    _name = "time_DV"
    _UID = ""
    _variable_label = "Time"
    _units = "s"
    _priority = 0
    _value_range = (0, 604800)  # don't record more than a week
    _value = 0

    _is_covariate = True

    # Initializes reference time.
    # The reference time usually denotes the beginning of an experiment trial.
    def __init__(self, *args, **kwargs):
        print(self._variable_label)
        super(DVTime, self).__init__(*args, **kwargs)
        print(self._variable_label)

    # Measure number of seconds relative to reference time
    def measure(self):

        value = time.time() - self._t0
        self.set_value(value)



import csv
import glob
import os
import shutil
import typing

import numpy as np
import torch
from torch import nn as nn

from autora.theorist.darts.model_search import Network
from autora.variable import ValueType


def create_output_file_name(
    file_prefix: str,
    log_version: int = None,
    weight_decay: float = None,
    k: int = None,
    seed: int = None,
    theorist: str = None,
) -> str:
    """
    Creates a file name for the output file of a theorist study.

    Arguments:
        file_prefix: prefix of the file name
        log_version: log version of the theorist run
        weight_decay: weight decay of the model
        k: number of nodes in the model
        seed: seed of the model
        theorist: name of the DARTS variant
    """

    output_str = file_prefix

    if theorist is not None:
        output_str += "_" + str(theorist)

    if log_version is not None:
        output_str += "_v_" + str(log_version)

    if weight_decay is not None:
        output_str += "_wd_" + str(weight_decay)

    if k is not None:
        output_str += "_k_" + str(k)

    if k is not None:
        output_str += "_s_" + str(seed)

    return output_str


def assign_slurm_instance(
    slurm_id: int,
    arch_weight_decay_list: typing.List,
    num_node_list: typing.List,
    seed_list: typing.List,
) -> typing.Tuple:
    """
    Determines the meta-search parameters based on the slum job id.

    Arguments:
        slurm_id: slurm job id
        arch_weight_decay_list: list of weight decay values
        num_node_list: list of number of nodes
        seed_list: list of seeds
    """

    seed_id = np.floor(
        slurm_id / (len(num_node_list) * len(arch_weight_decay_list))
    ) % len(seed_list)
    k_id = np.floor(slurm_id / (len(arch_weight_decay_list))) % len(num_node_list)
    weight_decay_id = slurm_id % len(arch_weight_decay_list)

    return (
        arch_weight_decay_list[int(weight_decay_id)],
        int(num_node_list[int(k_id)]),
        int(seed_list[int(seed_id)]),
    )


def get_loss_function(outputType: ValueType):
    """
    Returns the loss function for the given output type of a dependent variable.

    Arguments:
        outputType: output type of the dependent variable
    """

    dataSets = {
        ValueType.REAL: nn.MSELoss(),
        ValueType.PROBABILITY: sigmid_mse,
        ValueType.PROBABILITY_SAMPLE: sigmid_mse,
        ValueType.PROBABILITY_DISTRIBUTION: cross_entropy,
        ValueType.CLASS: nn.CrossEntropyLoss(),
        ValueType.SIGMOID: sigmid_mse,
    }

    return dataSets.get(outputType, nn.MSELoss)


def get_output_format(outputType: ValueType):
    """
    Returns the output format (activation function of the final output layer)
    for the given output type of a dependent variable.

    Arguments:
        outputType: output type of the dependent variable
    """
    dataSets = {
        ValueType.REAL: nn.Identity(),
        ValueType.PROBABILITY: nn.Sigmoid(),
        ValueType.PROBABILITY_SAMPLE: nn.Sigmoid(),
        ValueType.PROBABILITY_DISTRIBUTION: nn.Softmax(dim=1),
        ValueType.CLASS: nn.Softmax(dim=1),
        ValueType.SIGMOID: nn.Sigmoid(),
    }

    return dataSets.get(outputType, nn.MSELoss)


def get_output_str(outputType: ValueType) -> typing.Optional[str]:
    """
    Returns the output string for the given output type of a dependent variable.

    Arguments:
        outputType: output type of the dependent variable
    """

    dataSets = {
        ValueType.REAL: None,
        ValueType.PROBABILITY: "Sigmoid",
        ValueType.PROBABILITY_SAMPLE: "Sigmoid",
        ValueType.PROBABILITY_DISTRIBUTION: "Softmax",
        ValueType.CLASS: "Softmax",
        ValueType.SIGMOID: "Sigmoid",
    }

    return dataSets.get(outputType, nn.MSELoss)


def sigmid_mse(output: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """
    Returns the MSE loss for a sigmoid output.

    Arguments:
        output: output of the model
        target: target of the model
    """
    m = nn.Sigmoid()
    output = m(output)
    loss = torch.mean((output - target) ** 2)
    return loss


def compute_BIC(
    output_type: ValueType,
    model: torch.nn.Module,
    input: torch.Tensor,
    target: torch.Tensor,
) -> float:
    """
    Returns the Bayesian information criterion for a DARTS model.

    Arguments:
        output_type: output type of the dependent variable
        model: model to compute the BIC for
        input: input of the model
        target: target of the model
    """

    # compute raw model output
    classifier_output = model(input)

    # compute associated probability
    m = get_output_format(output_type)
    prediction = m(classifier_output).detach()

    k, _, _ = model.countParameters()  # for most likely architecture

    if output_type == ValueType.CLASS:
        target_flattened = torch.flatten(target.long())
        llik = 0
        for idx in range(len(target_flattened)):
            lik = prediction[idx, target_flattened[idx]]
            llik += np.log(lik)
        n = len(target_flattened)  # number of data points

        BIC = np.log(n) * k - 2 * llik
        BIC = BIC

    elif output_type == ValueType.PROBABILITY_SAMPLE:
        llik = 0
        for idx in range(len(target)):

            # fail safe if model doesn't produce probabilities
            if prediction[idx] > 1:
                prediction[idx] = 1
            elif prediction[idx] < 0:
                prediction[idx] = 0

            if target[idx] == 1:
                lik = prediction[idx]
            elif target[idx] == 0:
                lik = 1 - prediction[idx]
            else:
                raise Exception("Target must contain either zeros or ones.")
            llik += np.log(lik)
        n = len(target)  # number of data points

        BIC = np.log(n) * k - 2 * llik
        BIC = BIC[0]

    else:
        raise Exception(
            "BIC computation not implemented for output type "
            + str(ValueType.PROBABILITY)
            + "."
        )

    return BIC

    # old


def compute_BIC_AIC(
    soft_targets: np.array, soft_prediction: np.array, model: Network
) -> typing.Tuple:
    """
    Returns the Bayesian information criterion (BIC) as well as the
    Aikaike information criterion (AIC) for a DARTS model.

    Arguments:
        soft_targets: soft target of the model
        soft_prediction: soft prediction of the model
        model: model to compute the BIC and AIC for
    """

    lik = np.sum(
        np.multiply(soft_prediction, soft_targets), axis=1
    )  # likelihood of data given model
    llik = np.sum(np.log(lik))  # log likelihood
    n = len(lik)  # number of data points
    k, _, _ = model.countParameters()  # for most likely architecture

    BIC = np.log(n) * k - 2 * llik

    AIC = 2 * k - 2 * llik

    return BIC, AIC


def cross_entropy(pred: torch.Tensor, soft_targets: torch.Tensor) -> torch.Tensor:
    """
    Returns the cross entropy loss for a soft target.

    Arguments:
        pred: prediction of the model
        soft_targets: soft target of the model
    """
    # assuming pred and soft_targets are both Variables with shape (batchsize, num_of_classes),
    # each row of pred is predicted logits and each row of soft_targets is a discrete distribution.
    logsoftmax = nn.LogSoftmax(dim=1)
    return torch.mean(torch.sum(-soft_targets * logsoftmax(pred), 1))


class AvgrageMeter(object):
    """
    Computes and stores the average and current value.
    """

    def __init__(self):
        """
        Initializes the average meter.
        """
        self.reset()

    def reset(self):
        """
        Resets the average meter.
        """
        self.avg = 0
        self.sum = 0
        self.cnt = 0

    def update(self, val: float, n: int = 1):
        """
        Updates the average meter.

        Arguments:
            val: value to update the average meter with
            n: number of times to update the average meter
        """
        self.sum += val * n
        self.cnt += n
        self.avg = self.sum / self.cnt


def accuracy(
    output: torch.Tensor, target: torch.Tensor, topk: typing.Tuple = (1,)
) -> typing.List:
    """
    Computes the accuracy over the k top predictions for the specified values of k.

    Arguments:
        output: output of the model
        target: target of the model
        topk: values of k to compute the accuracy at
    """
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def count_parameters_in_MB(model: Network) -> int:
    """
    Returns the number of parameters for a model.

    Arguments:
        model: model to count the parameters for
    """
    return (
        np.sum(
            np.prod(v.size())
            for name, v in model.named_parameters()
            if "auxiliary" not in name
        )
        / 1e6
    )


def save(model: torch.nn.Module, model_path: str, exp_folder: str = None):
    """
    Saves a model to a file.

    Arguments:
        model: model to save
        model_path: path to save the model to
        exp_folder: general experiment directory to save the model to
    """
    if exp_folder is not None:
        os.chdir("exps")  # Edit SM 10/23/19: use local experiment directory
    torch.save(model.state_dict(), model_path)
    if exp_folder is not None:
        os.chdir("..")  # Edit SM 10/23/19: use local experiment directory


def load(model: torch.nn.Module, model_path: str):
    """
    Loads a model from a file.
    """
    model.load_state_dict(torch.load(model_path))


def create_exp_dir(
    path: str,
    scripts_to_save: typing.List = None,
    parent_folder: str = "exps",
    results_folder: str = None,
):
    """
    Creates an experiment directory and saves all necessary scripts and files.

    Arguments:
        path: path to save the experiment directory to
        scripts_to_save: list of scripts to save
        parent_folder: parent folder for the experiment directory
        results_folder: folder for the results of the experiment
    """
    os.chdir(parent_folder)  # Edit SM 10/23/19: use local experiment directory
    if not os.path.exists(path):
        os.mkdir(path)
    print("Experiment dir : {}".format(path))

    if results_folder is not None:
        try:
            os.mkdir(os.path.join(path, results_folder))
        except OSError:
            pass

    if scripts_to_save is not None:
        try:
            os.mkdir(os.path.join(path, "scripts"))
        except OSError:
            pass
        os.chdir("..")  # Edit SM 10/23/19: use local experiment directory
        for script in scripts_to_save:
            dst_file = os.path.join(
                parent_folder, path, "scripts", os.path.basename(script)
            )
            shutil.copyfile(script, dst_file)


def read_log_files(
    results_path: str, winning_architecture_only: bool = False
) -> typing.Tuple:
    """
    Reads the log files from an experiment directory and returns the results.

    Arguments:
        results_path: path to the experiment results directory
        winning_architecture_only: if True, only the winning architecture is returned
    """

    current_wd = os.getcwd()

    os.chdir(results_path)
    filelist = glob.glob("*.{}".format("csv"))

    model_name_list = list()
    loss_list = list()
    BIC_list = list()
    AIC_list = list()

    # READ LOG FILES

    print("Reading log files... ")
    for file in filelist:

        with open(file) as csvfile:
            readCSV = csv.reader(csvfile, delimiter=",")
            for row in readCSV:
                if winning_architecture_only is False or "sample0" in row[0]:
                    model_name_list.append(row[0])
                    loss_list.append(float(row[1]))
                    BIC_list.append(float(row[2].replace("[", "").replace("]", "")))
                    AIC_list.append(float(row[3].replace("[", "").replace("]", "")))

    os.chdir(current_wd)

    return (model_name_list, loss_list, BIC_list, AIC_list)


def get_best_fitting_models(
    model_name_list: typing.List,
    loss_list: typing.List,
    BIC_list: typing.List,
    topk: int,
) -> typing.Tuple:
    """
    Returns the topk best fitting models.

    Arguments:
        model_name_list: list of model names
        loss_list: list of loss values
        BIC_list: list of BIC values
        topk: number of topk models to return
    """

    topk_losses = sorted(zip(loss_list, model_name_list), reverse=False)[:topk]
    res = list(zip(*topk_losses))
    topk_losses_names = res[1]

    topk_BICs = sorted(zip(BIC_list, model_name_list), reverse=False)[:topk]
    res = list(zip(*topk_BICs))
    topk_BICs_names = res[1]

    return (topk_losses_names, topk_BICs_names)


def format_input_target(input, target, criterion):

    if isinstance(criterion, nn.CrossEntropyLoss):
        target = target.squeeze()

    return (input, target)



import numpy as np
import torch
import torch.nn.functional as F
from torch.autograd import Variable

from autora.theorist.darts.model_search import DARTSType, Network
from autora.theorist.darts.operations import isiterable


def _concat(xs) -> torch.Tensor:
    return torch.cat([x.view(-1) for x in xs])


class Architect(object):
    """
    A class representing a learner operating on the architecture weights of a DARTS model.
    This class handles training the weights associated with mixture operations
    (architecture weights).
    """

    def __init__(
        self,
        model: Network,
        arch_learning_rate: float,
        momentum: float,
        arch_weight_decay: float,
        arch_weight_decay_df: float = 0,
        arch_weight_decay_base: float = 0,
        fair_darts_loss_weight: float = 1,
    ):
        """
        Initializes the architecture learner.

        Arguments:
            model: a network model implementing the full DARTS model.
            arch_learning_rate: learning rate for the architecture weights
            momentum: momentum used in the Adam optimizer for architecture weights
            arch_weight_decay: general weight decay for the architecture weights
            arch_weight_decay_df: (weight decay applied to architecture weights in proportion
                to the number of parameters of an operation)
            arch_weight_decay_base: (a constant weight decay applied to architecture weights)
            fair_darts_loss_weight: (a regularizer that pushes architecture weights more toward
                zero or one in the fair DARTS variant)
        """
        # set parameters for architecture learning
        self.network_momentum = momentum
        self.network_weight_decay = arch_weight_decay
        self.network_weight_decay_df = arch_weight_decay_df
        self.arch_weight_decay_base = arch_weight_decay_base * model._steps
        self.fair_darts_loss_weight = fair_darts_loss_weight

        self.model = model
        self.lr = arch_learning_rate
        # architecture is optimized using Adam
        self.optimizer = torch.optim.Adam(
            self.model.arch_parameters(),
            lr=arch_learning_rate,
            betas=(0.5, 0.999),
            weight_decay=arch_weight_decay,
        )

        # initialize weight decay matrix
        self._init_decay_weights()

    def _init_decay_weights(self):
        """
        This function initializes the weight decay matrix. The weight decay matrix
        is subtracted from the architecture weight matrix on every learning step. The matrix
        specifies a weight decay which is proportional to the number of parameters used in an
        operation.
        """
        n_params = list()
        for operation in self.model.cells._ops[0]._ops:
            if isiterable(operation):
                n_params_total = (
                    1  # any non-zero operation is counted as an additional parameter
                )
                for subop in operation:
                    for parameter in subop.parameters():
                        if parameter.requires_grad is True:
                            n_params_total += parameter.data.numel()
            else:
                n_params_total = 0  # no operation gets zero parameters
            n_params.append(n_params_total)

        self.decay_weights = Variable(
            torch.zeros(self.model.arch_parameters()[0].data.shape)
        )
        for idx, param in enumerate(n_params):
            if param > 0:
                self.decay_weights[:, idx] = (
                    param * self.network_weight_decay_df + self.arch_weight_decay_base
                )
            else:
                self.decay_weights[:, idx] = param
        self.decay_weights = self.decay_weights
        self.decay_weights = self.decay_weights.data

    def _compute_unrolled_model(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        eta: float,
        network_optimizer: torch.optim.Optimizer,
    ):
        """
        Helper function used to compute the approximate architecture gradient.

        Arguments:
            input: input patterns
            target: target patterns
            eta: learning rate
            network_optimizer: optimizer used to updating the architecture weights

        Returns:
            unrolled_model: the unrolled architecture
        """
        loss = self.model._loss(input, target)
        theta = _concat(self.model.parameters()).data
        try:
            moment = _concat(
                network_optimizer.state[v]["momentum_buffer"]
                for v in self.model.parameters()
            ).mul_(self.network_momentum)
        except Exception:
            moment = torch.zeros_like(theta)
        dtheta = (
            _concat(torch.autograd.grad(loss, self.model.parameters())).data
            + self.network_weight_decay * theta
        )
        unrolled_model = self._construct_model_from_theta(
            theta.sub(eta, moment + dtheta)
        )
        return unrolled_model

    def step(
        self,
        input_valid: torch.Tensor,
        target_valid: torch.Tensor,
        network_optimizer: torch.optim.Optimizer,
        unrolled: bool,
        input_train: torch.Tensor = None,
        target_train: torch.Tensor = None,
        eta: float = 1,
    ):
        """
        Updates the architecture parameters for one training iteration

        Arguments:
            input_valid: input patterns for validation set
            target_valid: target patterns for validation set
            network_optimizer: optimizer used to updating the architecture weights
            unrolled: whether to use the unrolled architecture or not (i.e., whether to use
                the approximate architecture gradient or not)
            input_train: input patterns for training set
            target_train: target patterns for training set
            eta: learning rate for the architecture weights
        """

        # input_train, target_train only needed for approximation (unrolled=True)
        # of architecture gradient
        # when performing a single weigh update

        # initialize gradients to be zero
        self.optimizer.zero_grad()
        # use different backward step depending on whether to use
        # 2nd order approximation for gradient update
        if unrolled:  # probably using eta of parameter update here
            self._backward_step_unrolled(
                input_train,
                target_train,
                input_valid,
                target_valid,
                eta,
                network_optimizer,
            )
        else:
            self._backward_step(input_valid, target_valid)
        # move Adam one step
        self.optimizer.step()

    # backward step (using non-approximate architecture gradient, i.e., full training)
    def _backward_step(self, input_valid: torch.Tensor, target_valid: torch.Tensor):
        """
        Computes the loss and updates the architecture weights assuming full optimization
        of coefficients for the current architecture.

        Arguments:
            input_valid: input patterns for validation set
            target_valid: target patterns for validation set
        """
        if self.model.DARTS_type == DARTSType.ORIGINAL:
            loss = self.model._loss(input_valid, target_valid)
        elif self.model.DARTS_type == DARTSType.FAIR:
            loss1 = self.model._loss(input_valid, target_valid)
            loss2 = -F.mse_loss(
                torch.sigmoid(self.model.alphas_normal),
                0.5 * torch.ones(self.model.alphas_normal.shape, requires_grad=False),
            )  # torch.tensor(0.5, requires_grad=False)
            loss = loss1 + self.fair_darts_loss_weight * loss2
        else:
            raise Exception(
                "DARTS Type " + str(self.model.DARTS_type) + " not implemented"
            )

        loss.backward()

        # weight decay proportional to degrees of freedom
        for p in self.model.arch_parameters():
            p.data.sub_((self.decay_weights * self.lr))  # weight decay

    # backward pass using second order approximation
    def _backward_step_unrolled(
        self,
        input_train: torch.Tensor,
        target_train: torch.Tensor,
        input_valid: torch.Tensor,
        target_valid: torch.Tensor,
        eta: float,
        network_optimizer: torch.optim.Optimizer,
    ):
        """
        Computes the loss and updates the architecture weights using the approximate architecture
        gradient.

        Arguments:
            input_train: input patterns for training set
            target_train: target patterns for training set
            input_valid: input patterns for validation set
            target_valid: target patterns for validation set
            eta: learning rate
            network_optimizer: optimizer used to updating the architecture weights

        """

        # gets the model
        unrolled_model = self._compute_unrolled_model(
            input_train, target_train, eta, network_optimizer
        )

        if self.model.DARTS_type == DARTSType.ORIGINAL:
            unrolled_loss = unrolled_model._loss(input_valid, target_valid)
        elif self.model.DARTS_type == DARTSType.FAIR:
            loss1 = self.model._loss(input_valid, target_valid)
            loss2 = -F.mse_loss(
                torch.sigmoid(self.model.alphas_normal),
                torch.tensor(0.5, requires_grad=False),
            )
            unrolled_loss = loss1 + self.fair_darts_loss_weight * loss2
        else:
            raise Exception(
                "DARTS Type " + str(self.model.DARTS_type) + " not implemented"
            )

        unrolled_loss.backward()
        dalpha = [v.grad for v in unrolled_model.arch_parameters()]
        vector = [v.grad.data for v in unrolled_model.parameters()]
        implicit_grads = self._hessian_vector_product(vector, input_train, target_train)

        for g, ig in zip(dalpha, implicit_grads):
            g.data.sub_(eta, ig.data)

        for v, g in zip(self.model.arch_parameters(), dalpha):
            if v.grad is None:
                v.grad = Variable(g.data)
            else:
                v.grad.data.copy_(g.data)

    def _construct_model_from_theta(self, theta: torch.Tensor):
        """
        Helper function used to compute the approximate gradient update
        for the architecture weights.

        Arguments:
            theta: term used to compute approximate gradient update

        """
        model_new = self.model.new()
        model_dict = self.model.state_dict()

        params, offset = {}, 0
        for k, v in self.model.named_parameters():
            v_length = np.prod(v.size())
            params[k] = theta[offset : (offset + v_length)].view(v.size())
            offset += v_length

        assert offset == len(theta)
        model_dict.update(params)
        model_new.load_state_dict(model_dict)
        return model_new  # .cuda() # Edit SM 10/26/19: uncommented for cuda

    # second order approximation of architecture gradient (see Eqn. 8 from Liu et al, 2019)
    def _hessian_vector_product(
        self, vector: torch.Tensor, input: torch.Tensor, target: torch.Tensor, r=1e-2
    ):
        """
        Helper function used to compute the approximate gradient update
        for the architecture weights. It computes the hessian vector product outlined in Eqn. 8
        from Liu et al, 2019.

        Arguments:
            vector: input vector
            input: input patterns
            target: target patterns
            r: coefficient used to compute the hessian vector product

        """
        R = r / _concat(vector).norm()
        for p, v in zip(self.model.parameters(), vector):
            p.data.add_(R, v)
        loss = self.model._loss(input, target)
        grads_p = torch.autograd.grad(loss, self.model.arch_parameters())

        for p, v in zip(self.model.parameters(), vector):
            p.data.sub_(2 * R, v)
        loss = self.model._loss(input, target)
        grads_n = torch.autograd.grad(loss, self.model.arch_parameters())

        for p, v in zip(self.model.parameters(), vector):
            p.data.add_(R, v)

        # this implements Eqn. 8 from Liu et al. (2019)
        return [(x - y).div_(2 * R) for x, y in zip(grads_p, grads_n)]



from abc import abstractmethod

from tinkerforge.bricklet_industrial_analog_out_v2 import BrickletIndustrialAnalogOutV2
from tinkerforge.bricklet_industrial_dual_0_20ma_v2 import BrickletIndustrialDual020mAV2
from tinkerforge.bricklet_industrial_dual_analog_in_v2 import (
    BrickletIndustrialDualAnalogInV2,
)
from tinkerforge.ip_connection import IPConnection

from autora.variable import DV, IV, Variable


class TinkerforgeVariable(Variable):

    _variable_label = ""
    _UID = ""
    _priority = 0

    def __init__(
        self,
        variable_label="",
        UID="",
        name="",
        units="",
        priority="",
        value_range=(0, 1),
        type=float,
    ):

        super().__init__(
            name=name,
            value_range=value_range,
            units=units,
            type=type,
            variable_label=variable_label,
        )

        self._UID = UID
        self._priority = priority

    # Get priority of variable.
    # The priority is used to determine the sequence of variables to be measured or manipulated.
    def __get_priority__(self):
        return self._priority

    # Set priority of variable.
    # The priority is used to determine the sequence of variables to be measured or manipulated.
    def __set_priority__(self, priority):
        self._priority = priority

    @abstractmethod
    def clean_up(self):
        """Clean up measurement device."""
        pass

    @abstractmethod
    def disconnect(self):
        """Disconnect from up measurement device."""
        pass


class IVTF(IV, TinkerforgeVariable):
    def __init__(self, *args, **kwargs):
        IV.__init__(self, *args, **kwargs)
        TinkerforgeVariable.__init__(self, *args, **kwargs)


class DVTF(DV, TinkerforgeVariable):
    def __init__(self, *args, **kwargs):
        DV.__init__(self, *args, **kwargs)
        TinkerforgeVariable.__init__(self, *args, **kwargs)


class IVCurrent(IVTF):

    _name = "source_current"
    _UID = "MST"
    _variable_label = "Source Current"
    _units = "µA"
    _priority = 0
    _value_range = (0, 20000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    # Initializes Industrial Analog Out 2.0 device.
    def __init__(self, *args, **kwargs):

        self._ipcon = IPConnection()  # Create IP connection
        self._iao = BrickletIndustrialAnalogOutV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        super(IVCurrent, self).__init__(*args, **kwargs)

    # Clean up measurement device.
    def disconnect(self):

        self._iao.set_enabled(False)

        self._ipcon.disconnect()

    # Disable voltage output
    def stop(self):
        self._iao.set_enabled(False)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):
        self._iao.set_current(self.get_value())
        self._iao.set_enabled(True)

    def clean_up(self):
        self.stop()


class IVVoltage(IVTF):

    _variable_label = "Source Voltage"
    _UID = "MST"
    _name = "source_voltage"
    _units = "mV"
    _priority = 0
    _value_range = (0, 5000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    # Initializes Industrial Analog Out 2.0 device.
    def __init__(self, *args, **kwargs):

        self._ipcon = IPConnection()  # Create IP connection
        self._iao = BrickletIndustrialAnalogOutV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        super(IVVoltage, self).__init__(*args, **kwargs)

    # Clean up measurement device.
    def disconnect(self):

        self._iao.set_enabled(False)

        self._ipcon.disconnect()

    # Disable voltage output
    def stop(self):
        self._iao.set_enabled(False)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):
        self._iao.set_voltage(self.get_value())
        self._iao.set_enabled(True)

    def clean_up(self):
        self.stop()


class DVCurrent(DVTF):

    _name = "current0"
    _UID = "Hfg"
    _variable_label = "Current 0"
    _units = "mA"
    _priority = 0
    _value_range = (0, 2000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223
    channel = 0

    # Initializes Industrial Analog Out 2.0 device.
    def __init__(self, *args, **kwargs):

        super(DVCurrent, self).__init__(*args, **kwargs)

        self._ipcon = IPConnection()  # Create IP connection
        self._id020 = BrickletIndustrialDual020mAV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        if self._name == "current1":
            self.channel = 1
        else:
            self.channel = 0

    # Clean up measurement device.
    def disconnect(self):

        self._ipcon.disconnect()

    # Waits until specified time has passed relative to reference time
    def measure(self):
        current = self._id020.get_current(self.channel)
        self.set_value(current / 1000000.0)


class DVVoltage(DVTF):

    _name = "voltage0"
    _UID = "MjY"
    _variable_label = "Voltage 0"
    _units = "mV"
    _priority = 0
    _value_range = (-3500, 3500)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    channel = 0

    # Initializes Industrial Analog Out 2.0 device.
    def __init__(self, *args, **kwargs):

        super(DVVoltage, self).__init__(*args, **kwargs)

        self._ipcon = IPConnection()  # Create IP connection
        self._idai = BrickletIndustrialDualAnalogInV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        if self._name == "voltage1":
            self.channel = 1
        else:
            self.channel = 0

    # Clean up measurement device.
    def disconnect(self):
        self._ipcon.disconnect()

    # Waits until specified time has passed relative to reference time
    def measure(self):
        value = self._idai.get_voltage(self.channel)
        self.set_value(value)



import typing
from collections import namedtuple

import torch
import torch.nn as nn

Genotype = namedtuple("Genotype", "normal normal_concat")


def isiterable(p_object: typing.Any) -> bool:
    """
    Checks if an object is iterable.

    Arguments:
        p_object: object to be checked
    """
    try:
        iter(p_object)
    except TypeError:
        return False
    return True


def get_operation_label(
    op_name: str, params_org: typing.List, decimals: int = 4
) -> str:
    """
    Returns a complete string describing a DARTS operation.

    Arguments:
        op_name: name of the operation
        params_org: original parameters of the operation
        decimals: number of decimals to be used for converting the parameters into string format
    """
    params = params_org.copy()

    format_string = "{:." + "{:.0f}".format(decimals) + "f}"

    classifier_str = ""
    if op_name == "classifier":
        value = params[0]
        classifier_str = format_string.format(value) + " * x"

        return classifier_str

    if op_name == "classifier_concat":
        classifier_str = "x.*("
        for param_idx, param in enumerate(params):

            if param_idx > 0:
                classifier_str = classifier_str + " .+("

            if isiterable(param.tolist()):
                for value_idx, value in enumerate(param.tolist()):
                    if value_idx < len(param) - 1:
                        classifier_str = (
                            classifier_str + format_string.format(value) + " + "
                        )
                    else:
                        classifier_str = (
                            classifier_str + format_string.format(value) + ")"
                        )

            else:
                classifier_str = classifier_str + format_string.format(param) + ")"

        return classifier_str

    num_params = len(params)
    params.extend([0, 0, 0])

    if num_params == 1:  # without bias
        labels = {
            "none": "",
            "linear": str(format_string.format(params[0])) + " * x",
            "relu": "ReLU(x)",
            "lin_relu": "ReLU(" + str(format_string.format(params[0])) + " * x)",
            # 'sigmoid': '1/(1+e^(-x))',
            "sigmoid": "logistic(x)",
            # 'lin_sigmoid': '1/(1+e^(-' + str(format_string.format(params[0])) + ' * x))',
            "lin_sigmoid": "logistic(" + str(format_string.format(params[0])) + " * x)",
            "add": "+ x",
            "subtract": "- x",
            "mult": str(format_string.format(params[0])) + " * x",
            # 'exp': 'e^(' + str(format_string.format(params[0])) + ' * x)',
            "exp": "exp(" + str(format_string.format(params[0])) + " * x)",
            "1/x": "1 / (" + str(format_string.format(params[0])) + " * x)",
            "ln": "ln(" + str(format_string.format(params[0])) + " * x)",
            "classifier": classifier_str,
        }
    else:  # with bias
        labels = {
            "none": "",
            "linear": str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1])),
            "relu": "ReLU(x)",
            "lin_relu": "ReLU("
            + str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1]))
            + ")",
            # 'sigmoid': '1/(1+e^(-x))',
            "sigmoid": "logistic(x)",
            "lin_sigmoid": "logistic("
            + str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1]))
            + ")",
            "add": "+ x",
            "subtract": "- x",
            "mult": str(format_string.format(params[0])) + " * x",
            "exp": "exp("
            + str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1]))
            + ")",
            "1/x": "1 / ("
            + str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1]))
            + ")",
            "ln": "ln("
            + str(format_string.format(params[0]))
            + " * x + "
            + str(format_string.format(params[1]))
            + ")",
            "classifier": classifier_str,
        }

    return labels.get(op_name, "")


class Identity(nn.Module):
    """
    A pytorch module implementing the identity function.
    """

    def __init__(self):
        """
        Initializes the identify function.
        """
        super(Identity, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the identity function.

        Arguments:
            x: input tensor
        """
        return x


class NegIdentity(nn.Module):
    """
    A pytorch module implementing the inverse of an identity function.
    """

    def __init__(self):
        """
        Initializes the inverse of an identity function.
        """
        super(NegIdentity, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the inverse of an identity function.

        Arguments:
            x: input tensor
        """
        return -x


class Exponential(nn.Module):
    """
    A pytorch module implementing the exponential function.
    """

    def __init__(self):
        """
        Initializes the exponential function.
        """
        super(Exponential, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the exponential function.

        Arguments:
            x: input tensor
        """
        return torch.exp(x)


class NatLogarithm(nn.Module):
    """
    A pytorch module implementing the natural logarithm function.
    """

    def __init__(self):
        """
        Initializes the natural logarithm function.
        """
        super(NatLogarithm, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the natural logarithm function.

        Arguments:
            x: input tensor
        """
        # make sure x is in domain of natural logarithm
        mask = x.clone()
        mask[(x <= 0.0).detach()] = 0
        mask[(x > 0.0).detach()] = 1

        epsilon = 1e-10
        result = torch.log(nn.functional.relu(x) + epsilon) * mask

        return result


class MultInverse(nn.Module):
    """
    A pytorch module implementing the multiplicative inverse.
    """

    def __init__(self):
        """
        Initializes the multiplicative inverse.
        """
        super(MultInverse, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the multiplicative inverse.

        Arguments:
            x: input tensor
        """
        return torch.pow(x, -1)


class Zero(nn.Module):
    """
    A pytorch module implementing the zero operation (i.e., a null operation). A zero operation
    presumes that there is no relationship between the input and output.
    """

    def __init__(self, stride):
        """
        Initializes the zero operation.
        """
        super(Zero, self).__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the zero operation.

        Arguments:
            x: input tensor
        """
        if self.stride == 1:
            return x.mul(0.0)
        return x[:, :, :: self.stride, :: self.stride].mul(0.0)


class Softplus(nn.Module):
    r"""
    A pytorch module implementing the softplus function:

    $$
    \operatorname{Softplus}(x) = \frac{1}{β} \operatorname{log} \left( 1 + e^{β x} \right)
    $$
    """

    # This docstring is a raw-string (it starts `r"""` rather than `"""`)
    # so backslashes need not be escaped

    def __init__(self):
        """
        Initializes the softplus function.
        """
        super(Softplus, self).__init__()
        # self.beta = nn.Linear(1, 1, bias=False)
        self.beta = nn.Parameter(torch.ones(1))
        # elf.softplus = nn.Softplus(beta=self.beta)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the softplus function.

        Arguments:
            x: input tensor
        """
        y = torch.log(1 + torch.exp(self.beta * x)) / self.beta
        # y = self.softplus(x)
        return y


class Softminus(nn.Module):
    """
    A pytorch module implementing the softminus function: Softminus(x) = x- log(1+exp(β∗x)).

    $$
    \\operatorname{Softminus}(x) = x - \\operatorname{log} \\left( 1 + e^{β x} \\right)
    $$
    """

    # This docstring is a normal string, so backslashes need to be escaped

    def __init__(self):
        """
        Initializes the softminus function.
        """
        super(Softminus, self).__init__()
        # self.beta = nn.Linear(1, 1, bias=False)
        self.beta = nn.Parameter(torch.ones(1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the softminus function.

        Arguments:
            x: input tensor
        """
        y = x - torch.log(1 + torch.exp(self.beta * x)) / self.beta
        return y


# defines all the operations. affine is turned off for cuda (optimization prposes)
OPS = {
    "none": Zero(1),
    "linear": nn.Sequential(nn.Linear(1, 1, bias=True)),
    "relu": nn.Sequential(
        nn.ReLU(inplace=False),
    ),
    "lin_relu": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        nn.ReLU(inplace=False),
    ),
    "sigmoid": nn.Sequential(
        nn.Sigmoid(),
    ),
    "lin_sigmoid": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        nn.Sigmoid(),
    ),
    "add": nn.Sequential(Identity()),
    "subtract": nn.Sequential(NegIdentity()),
    "mult": nn.Sequential(
        nn.Linear(1, 1, bias=False),
    ),
    "exp": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        Exponential(),
    ),
    "1/x": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        MultInverse(),
    ),
    "ln": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        NatLogarithm(),
    ),
    "softplus": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        Softplus(),
    ),
    "softminus": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        Softminus(),
    ),
}

# this is the list of primitives actually used,
# and it should be a set of names contained in the OPS dictionary
PRIMITIVES = [
    "none",
    "add",
    "subtract",
    "linear",
    "lin_sigmoid",
    "mult",
    "lin_relu",
]

# make sure that every primitive is in the OPS dictionary
for name in PRIMITIVES:
    assert name in OPS



import numpy as np
import torch
from torch.utils.data import Dataset


class DARTSDataset(Dataset):
    def __init__(self, input_data: torch.tensor, output_data: torch.tensor):
        assert input_data.shape[0] == output_data.shape[0]
        self.input_data = input_data
        self.output_data = output_data

    def __len__(self, experiment_id=None):
        return self.input_data.shape[0]

    def __getitem__(self, idx):
        input_tensor = self.input_data[idx]
        output_tensor = self.output_data[idx]
        return input_tensor, output_tensor


def darts_dataset_from_ndarray(input_data: np.ndarray, output_data: np.ndarray):

    obj = DARTSDataset(
        torch.tensor(input_data, dtype=torch.get_default_dtype()),
        torch.tensor(output_data, dtype=torch.get_default_dtype()),
    )
    return obj



import typing

from graphviz import Digraph

from autora.theorist.darts.operations import Genotype, get_operation_label


def plot(
    genotype: Genotype,
    filename: str,
    file_format: str = "pdf",
    view_file: bool = None,
    full_label: bool = False,
    param_list: typing.Tuple = (),
    input_labels: typing.Tuple = (),
    out_dim: int = None,
    out_fnc: str = None,
):
    """
    Generates a graphviz plot for a DARTS model based on the genotype of the model.

    Arguments:
        genotype: the genotype of the model
        filename: the filename of the output file
        file_format: the format of the output file
        view_file: if True, the plot will be displayed in a window
        full_label: if True, the labels of the nodes will be the full name of the operation
            (including the coefficients)
        param_list: a list of parameters to be included in the labels of the nodes
        input_labels: a list of labels to be included in the input nodes
        out_dim: the number of output nodes of the model
        out_fnc: the (activation) function to be used for the output nodes
    """

    g = darts_model_plot(
        genotype=genotype,
        full_label=full_label,
        param_list=param_list,
        input_labels=input_labels,
        out_dim=out_dim,
        out_fnc=out_fnc,
    )

    if view_file is None:
        if file_format == "pdf":
            view_file = True
        else:
            view_file = False

    g.render(filename, view=view_file, format=file_format)


def darts_model_plot(
    genotype: Genotype,
    full_label: bool = False,
    param_list: typing.Sequence = (),
    input_labels: typing.Sequence = (),
    out_dim: int = None,
    out_fnc: str = None,
    decimals_to_display: int = 2,
) -> Digraph:
    """
    Generates a graphviz plot for a DARTS model based on the genotype of the model.

    Arguments:
        genotype: the genotype of the model
        full_label: if True, the labels of the nodes will be the full name of the operation
            (including the coefficients)
        param_list: a list of parameters to be included in the labels of the nodes
        input_labels: a list of labels to be included in the input nodes
        out_dim: the number of output nodes of the model
        out_fnc: the (activation) function to be used for the output nodes
        decimals_to_display: number of decimals to include in parameter values on plot
    """

    format_string = "{:." + "{:.0f}".format(decimals_to_display) + "f}"

    graph = Digraph(
        edge_attr=dict(fontsize="20", fontname="times"),
        node_attr=dict(
            style="filled",
            shape="rect",
            align="center",
            fontsize="20",
            height="0.5",
            width="0.5",
            penwidth="2",
            fontname="times",
        ),
        engine="dot",
    )
    graph.body.extend(["rankdir=LR"])

    for input_node in input_labels:
        graph.node(input_node, fillcolor="#F1EDB9")  # fillcolor='darkseagreen2'
    # assert len(genotype) % 2 == 0

    # determine number of steps (intermediate nodes)
    steps = 0
    for op, j in genotype:
        if j == 0:
            steps += 1

    for i in range(steps):
        graph.node(str(i), fillcolor="#BBCCF9")  # fillcolor='lightblue'

    params_counter = 0
    n = len(input_labels)
    start = 0
    for i in range(steps):
        end = start + n
        print(start, end)
        # for k in [2*i, 2*i + 1]:
        for k in range(
            start, end
        ):  # adapted this iteration from get_genotype() in model_search.py
            print(genotype, k)
            op, j = genotype[k]
            if j < len(input_labels):
                u = input_labels[j]
            else:
                u = str(j - len(input_labels))
            v = str(i)
            params_counter = k
            if op != "none":
                op_label = op
                if full_label:
                    params = param_list[
                        start + j
                    ]  # note: genotype order and param list order don't align
                    op_label = get_operation_label(
                        op, params, decimals=decimals_to_display
                    )
                    graph.edge(u, v, label=op_label, fillcolor="gray")
                else:
                    graph.edge(
                        u,
                        v,
                        label="(" + str(j + start) + ") " + op_label,
                        fillcolor="gray",
                    )  # '(' + str(k) + ') '
        start = end
        n += 1

    # determine output nodes

    out_nodes = list()
    if out_dim is None:
        out_nodes.append("out")
    else:
        biases = None
        if full_label:
            params = param_list[params_counter + 1]
            if len(params) > 1:
                biases = params[1]  # first node contains biases

        for idx in range(out_dim):
            out_str = ""
            # specify node ID
            if out_fnc is not None:
                out_str = out_str + out_fnc + "(r_" + str(idx)
            else:
                out_str = "(r_" + str(idx)

            if out_dim == 1:
                if out_fnc is not None:
                    out_str = "P(detected) = " + out_fnc + "(x"
                else:
                    # out_str = 'dx_1 = (x'
                    out_str = "P_n = (x"

            # if available, add bias
            if biases is not None:
                out_str = out_str + " + " + format_string.format(biases[idx]) + ")"
            else:
                out_str = out_str + ")"

            # add node
            graph.node(out_str, fillcolor="#CBE7C7")  # fillcolor='palegoldenrod'
            out_nodes.append(out_str)

    for i in range(steps):
        if full_label:
            params_org = param_list[params_counter + 1 + i]  # count from k
            for out_idx, out_str in enumerate(out_nodes):
                params = list()
                params.append(params_org[0][out_idx])
                op_label = get_operation_label(
                    "classifier", params, decimals=decimals_to_display
                )
                graph.edge(str(i), out_str, label=op_label, fillcolor="gray")
        else:
            for out_idx, out_str in enumerate(out_nodes):
                graph.edge(str(i), out_str, label="linear", fillcolor="gray")

    return graph
